%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FEASIBLE PERIODIC MOTION PRIMITIVES FOR CHOREOGRAPHED 
% QUADROCOPTER FLIGHT (submitted to ACC 2010)
% 
% **Calculating feasible parameter sets for side-to-side motion and 
% circlular motion** (Section V) 
%
% Angela Schoellig, Sept 2010 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function feasibilityData()

clear all
close all
clc


%% Choose between side-to-side motion (traj = 1) or circle (traj = 2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
traj = 1;
% define file name: 
% a .mat-file will be created at the end of the program using the provided 
% file name, the .mat-file saves the feasible parameter sets for the chosen 
% motion
filename = 'sidetoside_new';


%% Quadrocopter parameter (see Table 1 in the paper)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% If not mentioned otherwise, we use meter (m), second (s) and kilogram (kg) 
% as respective units. 

g = 9.81; % gravitational constant
I = zeros(3); % inertia matrix (assumed diagonal)
I(1,1) = 0.0023;
I(2,2) = 0.0023; 
I(3,3) = 0.0046;
k = 0.016; % motor constant, drag/thrust factor
m = 0.468; % vehicle mass
L = 0.17;  % vehicle arm length (=length from center of gravitiy to propellor)
fmin = 0.17; % minimum force of a motor normalized by the vehicle mass (in m/s^2)
fmax = 6; % maximum force of a motor normalized by the vehicle mass (in m/s^2)

% create a struct
PARAM.g = g;
PARAM.I =I;
PARAM.k = k;
PARAM.m = m;
PARAM.L = L;
PARAM.fmin = fmin;
PARAM.fmax = fmax;


%% Define parameter sets that are evaluated
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% minimum and maximum amplitude
Amin = 0.001;
Amax = 5;

% minimum and maximum frequency (see eq.(24) and (34))
Wmax = min(sqrt(sqrt((4*fmax)^2-g^2)/Amin),25);
Wmin = sqrt(sqrt((4*fmax)^2-g^2)/Amax);

% step length by which we increase the frequency and amplitude, respectively 
dW = .01;
dA= 0.001;

% resolution for numerical solution of dynamic equations: 
% number of points in one period (T =2*pi/W)
res = 1000; 


%% Calculate maximum feasible amplitude for a given frequency 
% based on collective thrust limits (Section IV.A)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

WVec = Wmin:dW:Wmax;
collA = getCollectiveA(WVec,PARAM);


%% Calculate maximum feasible amplitude for a given frequency 
% based on the thrust limits of a single motor (Section IV.B)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% variable storing the first (smallest) infeasible amplitude for each frequency
bounds = zeros(length(Wmin:dW:Wmax), 2);

% counter 
countW = 0;

% starting value for determing the smallest infeasible amplitude
A_old = inf;

% go through the frequencies
for W = Wmin:dW:Wmax
    
    % shows progress of computation
    countW = countW +1
    
    % calculate the largest feasible amplitude for the given frequency 
    % only considering the collective thrust limit 
    A = min(getCollectiveA(W,PARAM),A_old);
    
    % IMPORTANT: we start from A (which most likely is not feasible when 
    % considering individual motor limits) and decrease the amplitude until 
    % we first obtain a feasible amplitude 
    
    % boolean telling us if we crossed to border from infeasible to
    % feasible region
    boundscheck = 0;

    while (boundscheck ==0)&&(A>=Amin)
        
        % reduce amplitude
        A = A-dA;
        
        % returns the motor commands [f1;f2;f3;f4]; f1,f2, etc are row
        % vectors representing the motor command trajectories (over time)
        % from 0 to T.
        if traj == 1
            F = calculateMotorCmd_Side(A,W,PARAM,res);
        else
            F = calculateMotorCmd_Circle(A,W,PARAM,res);
        end
        
        % check single-motor thrust limits
        % is true if thrust limits are satisfied, finds the first feasible
        % amplitude
        if (sum(max(F')> fmax)==0)&&(sum(min(F')< fmin)==0)
            % save lowest infeasible amplitude, go one step up
            bounds(countW, 1:2) = [A+dA,1];
            boundscheck = 1;
            A_old = A+dA;
        end
    end
end

%% Basic plot of results
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
figure()
plot(WVec, [bounds(:,1)'; collA] )
xlabel('omega')
ylabel('amplitude')
if traj==1
    title('Side-to-Side')
else
    title('Circle')
end

grid on
         
findfigs

% save values
save ([filename,'.mat'], 'WVec', 'bounds', 'collA', 'PARAM', 'Amin', ...
    'Amax', 'dW', 'dA', 'Wmax', 'Wmin', 'res') 

end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Side-to-side motion
% function returning the motor commands motor_cmd = [f1;f2;f3;f4] for a 
% given amplitude and frequency; f1,f2, etc are row vectors representing 
% the motor command trajectories (over time)from 0 to T
function motor_cmd = calculateMotorCmd_Side(A,W,PARAM,res)
    
    freq = W/2/pi;
    t = (0:1/(res*freq):1/freq);
    
    %parameters
    g= PARAM.g;
    Iy = PARAM.I(2,2);
    m = PARAM.m;
    L = PARAM.L;
    beta=atan(-A*W^2/g * cos(W.*t));
    f = g./cos(beta);

    % from Mathematica, see eq. (28)
    betaDotDot = (2*A*g*W^4.*cos(W.*t).*(2*g^2 + 3*A^2*W^4 - A^2*W^4. *cos(2*W.*t)))./(2*g^2 + ...
        A^2*W^4 + A^2*W^4.*cos(2*W.*t)).^2;
    % from Mathematica, see eq. (30) and (31)
    motor_cmd =[ 0.5*(f./2+ Iy/m/L.*betaDotDot) ; f./4 ; 0.5*(f./2- Iy/m/L.*betaDotDot); f./4 ];

end

%% Circular motion
% function returning the motor commands motor_cmd = [f1;f2;f3;f4] for a 
% given amplitude and frequency; f1,f2, etc are row vectors representing 
% the motor command trajectories (over time)from 0 to T
function motor_cmd = calculateMotorCmd_Circle(A,W,PARAM,res)
    freq = W/2/pi;

    % parameters
    g= PARAM.g;
    I = PARAM.I;
    Iy = PARAM.I(2,2);
    Iz = PARAM.I(3,3);
    Ix = PARAM.I(1,1);
    m = PARAM.m;
    L = PARAM.L;
    k = PARAM.k;
    
    % short hands
    d = W^2*A/g;
    f = sqrt(g^2+W^4*A^2);
    
    motor_cmd = zeros(4,length(0:1/(res*freq):1/freq));
    count = 0;

    for t = (0:1/(res*freq):1/freq)
        count = count+1;
        % from Mathematica
        wX = (d*W *cos(W*t))/(sqrt(1 + d^2)* sqrt(1 - (d^2 *sin(W*t)^2)/(1 + d^2)));
        wY = (d*W *sin(W*t)* sqrt(1 - (d^2 *sin(W*t)^2)/(1 + d^2)))/(1 + d^2* cos(W*t)^2);
        wZ = -((d^2 *W *sin(W.*t)^2)/(sqrt(1 + d^2)* (1 + d^2 *cos(W.*t)^2)));
        wVec = [wX; wY; wZ];
        wXDot = -((2 *sqrt(2)* d *W^2 *sin(W.*t))/((1 + d^2)^(3/2)*((2 + d^2 + d^2 *cos(2* t *W))/...
            (1 + d^2))^(3/2)));
        wYDot = ((2 *sqrt(2)* d *W^2 *cos(W.*t))/((1 + d^2)*((2 + d^2 + d^2 *cos(2* t *W))/...
            (1 + d^2))^(3/2)));
        wZDot = -((d^2 *sqrt(1 + d^2)* W^2 *sin(2*t*W))/(1 + d^2 *cos(t*W)^2)^2);
        wDotVec = [wXDot; wYDot; wZDot];
        
        % compute M = inverse of S = m.*[0 L 0 -L; -L 0 L 0; k -k k -k; 1 1
        % 1 1], cf. eq. (11) and (7) with mass-normalized motor forces f1, f2, etc.
        M = [ 0, -1/(2*L*m),  1/(4*k*m), 1/(4*m);
            1/(2*L*m),          0, -1/(4*k*m), 1/(4*m);
          0,  1/(2*L*m),  1/(4*k*m), 1/(4*m);
         -1/(2*L*m),          0, -1/(4*k*m), 1/(4*m)];
     
        % from eq. (11) and (7)
        motor_cmd(:,count) = M * [I*wDotVec + cross(wVec,I*wVec); f*m];
    end


end

% returns the maximum amplitude for a given frequency based on the
% collective thrust limits, see eq. (24) and (34)
function A = getCollectiveA(W,PARAM)
    A = sqrt((4*PARAM.fmax)^2-PARAM.g^2)./W.^2;
end